import pytest

from constant_sorrow import constants


def test_establishing_a_constant():
    # You get a constant by just picking an all-caps name and importing it.
    from constant_sorrow.constants import THIS_IS_A_VALID_CONSTANT

    # That name didn't exist before - you literally just make up
    # an all-caps name and import it.

    # But you can't make a constant in lower case.
    with pytest.raises(ValueError):
        constants.this_is_not_the_right_case_for_a_constant


def test_different_constants_are_unequal():
    # Constants with different names are not equal to each other by default.
    assert constants.ONE_THING != constants.ANOTHER_THING


def test_same_constants_are_identical():
    # However, constants with the same name are both equal and identical to one another.
    # They are literally the same object.
    from constant_sorrow.constants import SAME_THING
    assert constants.SAME_THING == SAME_THING
    assert constants.SAME_THING is SAME_THING


def test_set_representation():
    # Merely using a constant as a special value flag might not be enough.
    # You might want to represent it for the purposes of, for example,
    # sending it over the wire as bytes.
    from constant_sorrow.constants import AFRICAN_SWALLOW
    AFRICAN_SWALLOW(b"non-migratory")
    assert bytes(AFRICAN_SWALLOW) == b"non-migratory"


def test_bytes_representation_default():
    """By default, constants are represented as a SHA512 hash of their name, truncated to 8 bytes."""
    another_constant = constants.ANOTHER_CONSTANT
    assert len(bytes(another_constant)) == 8
    assert bytes(another_constant) == constants.ANOTHER_CONSTANT


def test_cast_representation():
    constants.FOURTEEN(b"14")

    assert int(constants.FOURTEEN) == 14
    assert bytes(constants.FOURTEEN) == b"14"
    assert str(constants.FOURTEEN) == "14"


def test_cant_change_representation():
    constants.DINGOS("a certain dingo")

    # We can't change the value once it is set.
    with pytest.raises(ValueError):
        constants.DINGOS("something else")

    # However setting the same value again is permitted.
    constants.DINGOS("a certain dingo")

    assert bytes(constants.DINGOS) == b"a certain dingo"


def test_bool_representation():
    # Unlike representing as bytes, you can't automatically represent as bool.
    with pytest.raises(TypeError):
        bool(constants.NO_KNOWN_BOOL)

    # You can either set the representation...
    constants.WITH_BOOL_FROM_REPR("non-empty strings are True, obviously.")
    assert bool(constants.WITH_BOOL_FROM_REPR) is True

    # Or you can specifically set a bool representation.
    constants.WITH_SET_BOOL.bool_value(False)
    assert bool(constants.WITH_SET_BOOL) is False

    # A set boolean value will take precedence over the representation.
    # (because it's a Constant - it'd be bizarre behavior for the bool representation to change)
    constants.WITH_SET_BOOL("this string is non-empty, but this constant is still False.")
    assert bool(constants.WITH_SET_BOOL) is False

    # You can set it to the same value again...
    constants.WITH_BOOL.bool_value(False)

    # But you can't change the bool value once set.
    with pytest.raises(ValueError):
        constants.WITH_BOOL.bool_value(True)

    # Also, we can't take the first constant above and set its bool value to False.
    # It was True before, because prior to it having a bool_value, it was represented
    # by a non-empty string.  So now, we can't change it to False.
    # After all - we want it to act like a constant.
    with pytest.raises(ValueError):
        constants.WITH_BOOL_FROM_REPR.bool_value(False)

    # We can, however, set it to True, because its current value is True.
    constants.WITH_BOOL_FROM_REPR.bool_value(True)
    assert bool(constants.WITH_BOOL_FROM_REPR) is True


def test_str_representation():
    # By default, Constants cast to str as their name:
    str(constants.BECAUSE_IT_IS_MY_NAME) == "BECAUSE_IT_IS_MY_NAME"

    # If a Constant is ever cast this way, its representation can't be changed to
    # an object whose string representation is different.
    # For example, this object that casts to the same string:

    class BecauseICannotHaveAnotherInMyLife:
        def __str__(self):
            return "BECAUSE_IT_IS_MY_NAME"

    constants.BECAUSE_IT_IS_MY_NAME(BecauseICannotHaveAnotherInMyLife())

    # ...but this one will not cast away its good name, sir:
    class AbigailWilliams:
        def __str__(self):
            return "Mylène Demongeot"

    str(constants.WINONA_RYDER)

    with pytest.raises(ValueError):
        constants.WINONA_RYDER(AbigailWilliams())

    # Note: it's possible to use an object which initially has a str repr that matches, but later changes.
    # This is almost certainly a bad idea, but if you're going to go that far, we trust you have a reason.
    class BecauseILieAndSignMyselfToLies:
        name = "JOHN PROCTOR"

        def __str__(self):
            return self.name

    constants.JOHN_PROCTOR(BecauseILieAndSignMyselfToLies())
    BecauseILieAndSignMyselfToLies.name = "Daniel Plainview"
    assert str(constants.JOHN_PROCTOR) == "Daniel Plainview"
    # ...but again, why?  This seems like a bad idea, but it's also not the place
    # of the Constant class to police the internal behavior of your classes.


def test_cant_set_attrs():
    # You can't set an attr on a constant.
    with pytest.raises(TypeError):
        constants.FISH_SLAPPING_DANCE.whatever = 4


def test_repr_as_str():
    # Constants are repr'd by their name; simple.
    from constant_sorrow.constants import HOLY_HAND_GRENADE
    assert repr(HOLY_HAND_GRENADE) == "HOLY_HAND_GRENADE"
    # If they have a representation set, that is added.
    HOLY_HAND_GRENADE("One, Two, Five!")
    assert repr(HOLY_HAND_GRENADE) == "HOLY_HAND_GRENADE (One, Two, Five!)"


def test_arithmetic():
    # Constants are summed by their representation.
    constants.FORTY_TWO(42)
    assert 80 + constants.FORTY_TWO == 80 + 42

    # Adding strings to a constant with a string representation yields a string.
    from constant_sorrow.constants import ROBERT
    ROBERT("Bob")
    assert ROBERT + " up and down in the water" == "Bob up and down in the water"

    # But adding bytes causes a cast to bytes.
    assert b"Sideshow " + ROBERT == b'Sideshow Bob'

    # Same with int.
    constants.THIRTY_SEVEN(b"37")
    assert constants.THIRTY_SEVEN + 10 == 47

    # sub, mul, and truediv work also.
    assert constants.THIRTY_SEVEN - 30 == 7
    assert 137 - constants.THIRTY_SEVEN == 100

    assert constants.THIRTY_SEVEN * 2 == 74
    assert 3 * constants.THIRTY_SEVEN == 111

    assert constants.THIRTY_SEVEN / 4 == 9.25
    assert 111 / constants.THIRTY_SEVEN == 3.0


def test_constants_can_be_used_as_ints():
    # A constant which is represented as an int can be used in most places
    # that an int can be used, even without casting first.
    constants.THREE(3)
    assert "humbug"[:constants.THREE] == "hum"


def test_constant_length():
    # By default, the length of a constant is the length of its name...
    assert len(constants.PEAR) == 4

    # ...but once a representation is set, then it will be that length.
    constants.PEAR("fruit")
    assert len(constants.PEAR) == 5


def test_iterable_constant():
    # Setup a new iterable constant of string literals
    names = ('Ron', 'Bob', 'Jerry', 'Phil', 'Keith', 'Donna')
    constants.MEMBERS(names)

    # The constant can be cast to a list...
    assert list(constants.MEMBERS) == list(names)

    # python 3.6 test
    # pigpen, *the_rest = list(constants.MEMBERS)
    # assert pigpen == 'Ron'
    # assert len(the_rest) == 5

    # ...unpacked into a function.
    def take_input_and_do_nothing_with_it(a, b, c, d, e, f): return True

    assert take_input_and_do_nothing_with_it(*constants.MEMBERS)

    # Hold a collection of constants...
    _weather = (constants.THUNDER('⛈'),
                constants.LIGHTNING('⚡'),
                constants.WIND('💨'),
                constants.RAIN('💧'))

    constants.WEATHER(_weather)

    def operate_on_input(a, b, c, d):
        assert a + b + c + d == '⛈⚡💨💧'
        return True

    # ...unpack and operate on the elements directly...
    assert operate_on_input(*constants.WEATHER)

    # ...or iterate over the constant
    assert ''.join(str(item) for item in constants.WEATHER) == '⛈⚡💨💧'


def test_use_methods_on_representation():
    # By default, accessing strange attributes with raise AttributeError.
    from constant_sorrow.constants import THE_OLD_MAN_THE_BOAT
    with pytest.raises(AttributeError):
        THE_OLD_MAN_THE_BOAT.upper()

    # If a representation is set, attributes other than ("__repr_content", "__bool_repr", "__name") pass through.
    THE_OLD_MAN_THE_BOAT("when Fred eats food gets thrown.")
    assert THE_OLD_MAN_THE_BOAT.upper() == 'WHEN FRED EATS FOOD GETS THROWN.'


def test_useful_name_in_exception_output():
    """
    Constants use a metaclass which uses the Constant name as the string representation of the class (not only the instance).
    This is useful for meaningful error messages.
    """
    try:
        constants.NORTHERN_RAILROAD[:7]
    except TypeError as e:
        assert e.args[0] == "'NORTHERN_RAILROAD' object is not subscriptable"
    else:
        pytest.fail("Expected this constant not to be subscriptable by default.")


def test_setting_documentation():
    from constant_sorrow.constants import NORTHERN_RAILROAD
    assert not NORTHERN_RAILROAD.__doc__

    new_doc = "Maybe I'll die upon this train."
    NORTHERN_RAILROAD.set_constant_documentation(new_doc)
    assert NORTHERN_RAILROAD.__doc__ == new_doc

    from constant_sorrow.constants import MY_OWN_TRUE_LOVER
    NORTHERN_RAILROAD.set_constant_documentation("I never expect to see you again.")

    assert NORTHERN_RAILROAD.__doc__ != MY_OWN_TRUE_LOVER.__doc__


import hashlib
import sys
from copy import deepcopy
from types import ModuleType

from . import _digest_length


def hash_and_truncate(constant):
    return hashlib.sha512(constant._Constant__name.encode()).digest()[:_digest_length]


class _Constant:
    __repr_content = None
    __bool_repr = None
    __uses_default_repr = True
    __has_been_stringified = False

    __doc__ = "Maybe your friends think this is just an instance; an object you'll never see any more."

    class OldKentucky(RuntimeError):
        pass

    def __init__(self, name):
        # Names must be upper case except that dunder names can be set.
        # They're used by a lot of IDE tooling, so let's not mess up a good thing.
        if not name.isupper():
            if not (name.startswith("__") and name.endswith("__")):
                raise ValueError(
                    "Use ALL_CAPS names for constants.  See https://www.python.org/dev/peps/pep-0008/#constants.")
        self.__name = name

    def __setattr__(self, key, value):
        if key in ("_Constant__repr_content", "_Constant__bool_repr", "_Constant__name", "_Constant__uses_default_repr",
                   "_Constant__has_been_stringified"):
            super().__setattr__(key, value)
        else:
            raise TypeError("Don't try to set values on a constant.  I mean, what's the point?")

    def __getattr__(self, item):
        try:
            return getattr(self.__repr_content, item)
        except AttributeError:
            raise AttributeError("Without a representation, you can't use {}.".format(item))

    def __bytes__(self):
        if type(self.__repr_content) == str:
            return self._cast_repr(bytes, encoding="utf-8")
        else:
            return self._cast_repr(bytes)

    def __int__(self):
        return self._cast_repr(int)

    def __str__(self):
        # Unless there's an explicit repr, we want the str value to be the name.
        if type(self.__repr_content) is None or self.__uses_default_repr:
            self.__has_been_stringified = True
            return self.__name
        if type(self.__repr_content) == bytes:
            return self._cast_repr(str, encoding="utf-8")
        else:
            return self._cast_repr(str)

    def __bool__(self):
        if self.__bool_repr is None:
            if self.__repr_content is None:
                raise TypeError("This constant can't currently be represented as a bool.")
            else:
                return bool(self.__repr_content)
        else:
            return self.__bool_repr

    def __repr__(self):
        if self.__repr_content is not None:
            return "{} ({})".format(self.__name, self.__repr_content)
        else:
            return self.__name

    def __add__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) + other

    def __radd__(self, other):
        return other + self._cast_to_other_object_type_or_bytes(other)

    def __sub__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) - other

    def __rsub__(self, other):
        return other - self._cast_to_other_object_type_or_bytes(other)

    def __mul__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) * other

    def __rmul__(self, other):
        return other * self._cast_to_other_object_type_or_bytes(other)

    def __truediv__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) / other

    def __rtruediv__(self, other):
        return other / self._cast_to_other_object_type_or_bytes(other)

    def __floordiv__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) // other

    def __rfloordiv__(self, other):
        return other // self._cast_to_other_object_type_or_bytes(other)

    def __gt__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) > other

    def __ge__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) >= other

    def __lt__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) < other

    def __le__(self, other):
        return self._cast_to_other_object_type_or_bytes(other) <= other

    def __eq__(self, other):
        try:
            for_comparison_sake = self._cast_to_other_object_type_or_bytes(other)
        except ValueError:  # Can't cast to the other type, so obviously this isn't equal.
            return False
        return for_comparison_sake == other

    def __hash__(self):
        return hash(self.__repr_content)

    def __call__(self, representation):
        representation_will_change = self.__repr_content is not None and self.__repr_content is not representation
        if representation_will_change:
            message = "Can't set representation to a different value once set - it was " \
                      "already set to {} when you tried to set it to {}"
            raise ValueError(message.format(self.__repr_content, representation))

        if self.__has_been_stringified:
            if not self.__name == str(representation):
                message = "This Constant has already been represented as the string {} and can't be changed to be represented by {}"
                raise ValueError(message.format(self.__name, str(representation)))

        elif self.__repr_content is representation:
            return self
        else:
            self.__uses_default_repr = False
            self.__repr_content = deepcopy(representation)

        return self

    def __index__(self):
        return int(self)

    def __len__(self):
        if self.__repr_content is not None:
            return len(self.__repr_content)
        else:
            return len(self.__name)

    def __iter__(self):
        for item in self.__repr_content:
            yield item

    @classmethod
    def set_constant_documentation(cls, doc):
        cls.__doc__ = doc

    @property
    def _sorrow_type(self):
        if self.__repr_content is None:
            raise self.OldKentucky
        repr_type = type(self.__repr_content)
        return repr_type

    def _cast_to_other_object_type_or_bytes(self, other):
        if type(other) in (bytes, int, str):
            # Cast to other object type if it's bytes, int, or str.
            caster = type(other)
        elif _Constant in other.__class__.__bases__:
            try:
                caster = other._sorrow_type
            except self.OldKentucky:
                caster = bytes
        else:
            caster = bytes
        return caster(self)

    def _cast_repr(self, caster, *args, **kwargs):
        """
        Will cast this constant with the provided caster, passing args and kwargs.

        If there is no registered representation, will hash the name using sha512 and use the first 8 bytes
        of the digest.
        """
        if self.__repr_content is None:
            self.__repr_content = hash_and_truncate(self)
            assert self.__uses_default_repr  # Sanity check: we are indeed using the default repr here.  If this has ever changed, something went wrong.

        return caster(self.__repr_content, *args, **kwargs)

    def bool_value(self, bool_value):
        if self.__repr_content is not None:
            if bool(self) is not bool(bool_value):
                raise ValueError("Based on the set representation, {} was previously {}; can't change to {}.".format(
                    self.__name,
                    bool(self),
                    bool(bool_value)))

        if self.__bool_repr is not None:
            if bool(self) is not bool(bool_value):
                raise ValueError("The specified bool value for {} was previously {}; can't change to {}.".format(
                    self.__name,
                    bool(self),
                    bool(
                        bool_value)))

        self.__bool_repr = bool(bool_value)
        return self


_constants_registry_by_name = {}
_constants_registry_by_hash = {}


class __ConstantFactory(ModuleType):

    def __getattr__(self, item):

        try:
            constant = _constants_registry_by_name[item.upper()]
        except KeyError:

            _constant_class = type(item, (_Constant,), {}) # The actual class of the constant we'll return.
            constant = _constant_class(item)
            _constants_registry_by_name[item.upper()] = constant
            _constants_registry_by_hash[hash_and_truncate(constant)] = constant

        return constant

sys.modules[__name__].__class__ = __ConstantFactory



from constant_sorrow.__about__ import __author__, __summary__, __title__, __version__
__all__ = ["__title__", "__summary__", "__version__", "__author__", ]


from bytestring_splitter import BytestringSplitter
_digest_length = 8

default_constant_splitter = key_splitter = BytestringSplitter((bytes, _digest_length))


def constant_or_bytes(possible_constant):
    from .constants import _Constant
    from .constants import _constants_registry_by_hash
    """
    Utility function for getting a constant (that has already been registered) from a serialized constant (ie, bytes of its hash)
    """
    if _Constant in possible_constant.__class__.__bases__:
        result = possible_constant
    else:
        bytes_of_possible_constant = bytes(possible_constant)
        try:
            constant = _constants_registry_by_hash[bytes_of_possible_constant]
            result = constant
        except KeyError:
            result = bytes_of_possible_constant
    return result


from __future__ import absolute_import, division, print_function

__all__ = [
    "__title__", "__summary__", "__version__", "__author__", "__email__", "__license__", "__copyright__", "__url__"
]

__title__ = "constant_sorrow"

__url__ = "https://github.com/nucypher/constantSorrow"

__summary__ = "I am the man of constant sorrow; I've seen special values as ints, all my days."

__version__ = "v0.1.0-alpha.7"

__author__ = "NuCypher"

__license__ = "Apache Software License v2.0"

__email__ = "dev@nucypher.com"

__copyright__ = 'Copyright (C) 2018 NuCypher'



import os
import sys

from setuptools import setup
from setuptools.command.install import install

BASE_DIR = os.path.dirname(__file__)

ABOUT = dict()
with open(os.path.join(BASE_DIR, "constant_sorrow", "__about__.py")) as f:
    exec(f.read(), ABOUT)


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != ABOUT['__version__']:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, ABOUT['__version__']
            )
            sys.exit(info)


INSTALL_REQUIRES = ['bytestring-splitter']
EXTRAS_REQUIRE = {'testing': ['pytest', 'bumpversion'],
                  'docs': ['sphinx', 'sphinx-autobuild']}

setup(name=ABOUT['__title__'],
      url=ABOUT['__url__'],
      version=ABOUT['__version__'],
      author=ABOUT['__author__'],
      author_email=ABOUT['__email__'],
      description=ABOUT['__summary__'],
      extras_require=EXTRAS_REQUIRE,
      install_requires=INSTALL_REQUIRES,
      setup_requires=['pytest-runner'],  # required for setup.py test
      packages=['constant_sorrow'],
      classifiers=[
          "Development Status :: 2 - Pre-Alpha",
          "Natural Language :: English",
          "Programming Language :: Python :: Implementation",
          "Programming Language :: Python :: 3 :: Only",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Topic :: Software Development :: Libraries :: Python Modules"
      ],
      python_requires='>=3',
      cmdclass={'verify': VerifyVersionCommand}
      )
